#pragma once

#include <cstdlib>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>

namespace BGTU::ComputerGraphicsLabWork::Utilities {
	class brezenham_line_iterable {
	private:
		RendererApi::PointI<2>::component_t start_arg, start_ret, end_arg;
		RendererApi::PointI<2>::component_t delta_ret;
		RendererApi::PointI<2>::component_t delta_error, error_threshold;
		bool swap_flag;
	public:
		inline brezenham_line_iterable(RendererApi::PointI<2>::component_t x1, RendererApi::PointI<2>::component_t y1, RendererApi::PointI<2>::component_t x2, RendererApi::PointI<2>::component_t y2) {
			RendererApi::PointI<2>::component_t delta_x = (x1 > x2) ? (x1 - x2) : (x2 - x1);
			RendererApi::PointI<2>::component_t delta_y = (y1 > y2) ? (y1 - y2) : (y2 - y1);
			if (delta_x > delta_y) {
				this->start_arg = x1;
				this->start_ret = y1;
				this->end_arg = x2;
				this->delta_error = (RendererApi::PointI<2>::component_t) (delta_y + 1);
				this->error_threshold = (RendererApi::PointI<2>::component_t) (delta_x + 1);
				this->delta_ret = (y2 > y1) ? 1 : -1;
				this->swap_flag = false;
			} else {
				this->start_arg = y1;
				this->start_ret = x1;
				this->end_arg = y2;
				this->delta_error = (RendererApi::PointI<2>::component_t) (delta_x + 1);
				this->error_threshold = (RendererApi::PointI<2>::component_t) (delta_y + 1);
				this->delta_ret = (x2 > x1) ? 1 : -1;
				this->swap_flag = true;
			}
		}

		struct point {
		public:
			RendererApi::PointI<2>::component_t x, y;

			inline point(RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) : x{x}, y{y} {}
		};

		class iterator {
		private:
			RendererApi::PointI<2>::component_t arg, ret, end_arg;
			RendererApi::PointI<2>::component_t delta_ret;
			RendererApi::PointI<2>::component_t error, delta_error, error_threshold;
			bool swap_flag;

			inline iterator(
					RendererApi::PointI<2>::component_t arg, RendererApi::PointI<2>::component_t ret, RendererApi::PointI<2>::component_t end_arg,
					RendererApi::PointI<2>::component_t delta_ret,
					RendererApi::PointI<2>::component_t delta_error, RendererApi::PointI<2>::component_t error_threshold,
					bool swap_flag
			) : arg{arg}, ret{ret}, end_arg{end_arg}, delta_ret{delta_ret}, error{0}, delta_error{delta_error}, swap_flag{swap_flag}, error_threshold{error_threshold} {}

			friend class brezenham_line_iterable;

		public:
			inline iterator &operator++() {
				this->arg++;
				this->error += this->delta_error;
				if (this->error >= this->error_threshold) {
					this->ret += this->delta_ret;
					this->error -= this->error_threshold;
				}
				return *this;
			}

			inline iterator operator++(int) {
				iterator next = *this;
				++(*this);
				return next;
			}

			inline point operator*() const {
				if (this->swap_flag)
					return point{this->ret, this->arg};
				else
					return point{this->arg, this->ret};
			}

			inline bool operator==(std::nullptr_t) const {
				return this->arg > this->end_arg;
			}

			inline bool operator!=(std::nullptr_t) const {
				return this->arg <= this->end_arg;
			}
		};

		[[nodiscard]] inline iterator begin() const {
			return iterator{
					this->start_arg, this->start_ret, this->end_arg,
					this->delta_ret,
					this->delta_error, this->error_threshold,
					this->swap_flag
			};
		}

		[[nodiscard]] inline std::nullptr_t end() const {
			return nullptr;
		}
	};
}