#pragma once

#include <climits>
#include <emmintrin.h>
#include <bgtu/computer_graphics_lab_work/renderer_api/color.hpp>


namespace BGTU::ComputerGraphicsLabWork::Utilities {
	inline RendererApi::Color apply_transparent_color(RendererApi::Color bg, RendererApi::Color::Transparent fg) {
#if defined(__AVX512F__)
		return _apply_transparent_color_avx512(bg, fg);
#elif 0 && defined(__AVX2__)
		return _apply_transparent_color_avx2(bg, fg);
#else
#if 1
		return RendererApi::Color{
				static_cast<RendererApi::Color::component_fast_t>(((static_cast<std::uint_fast16_t>(fg.red) * static_cast<std::uint_fast16_t>(fg.alpha) + static_cast<std::uint_fast16_t>(bg.red) * (static_cast<std::uint_fast16_t>(255) - static_cast<std::uint_fast16_t>(fg.alpha))) / 255)),
				static_cast<RendererApi::Color::component_fast_t>(((static_cast<std::uint_fast16_t>(fg.green) * static_cast<std::uint_fast16_t>(fg.alpha) + static_cast<std::uint_fast16_t>(bg.green) * (static_cast<std::uint_fast16_t>(255) - static_cast<std::uint_fast16_t>(fg.alpha))) / 255)),
				static_cast<RendererApi::Color::component_fast_t>(((static_cast<std::uint_fast16_t>(fg.blue) * static_cast<std::uint_fast16_t>(fg.alpha) + static_cast<std::uint_fast16_t>(bg.blue) * (static_cast<std::uint_fast16_t>(255) - static_cast<std::uint_fast16_t>(fg.alpha))) / 255)),
		};
#elif 1
		static auto calc = [&](RendererApi::Color::component_fast_t s, RendererApi::Color::component_fast_t o) {
			auto oo = static_cast<std::uint_fast16_t>(o) * static_cast<std::uint_fast16_t>(fg.alpha);
			auto ss = static_cast<std::uint_fast16_t>(s) * (static_cast<std::uint_fast16_t>(255) - static_cast<std::uint_fast16_t>(fg.alpha));
			return static_cast<std::uint_fast8_t >((oo + ss) / 256);
		};
		return RendererApi::Color{
				calc(bg.red, fg.red),
				calc(bg.green, fg.green),
				calc(bg.blue, fg.blue)
		};
#else
		static __m128i vec_256 = _mm_set_epi16(256, 256, 256, 0, 0, 0, 0, 0);
		__m128i vec_bg = _mm_set_epi16(bg.red, bg.green, bg.blue, 0, 0, 0, 0, 0);
		__m128i vec_fg = _mm_set_epi16(fg.red, fg.green, fg.blue, 0, 0, 0, 0, 0);
		__m128i vec_alpha = _mm_set_epi16(fg.alpha, fg.alpha, fg.alpha, 0, 0, 0, 0, 0);

		union {
			__m128i packed;
			std::uint16_t unpacked[3];
		} raw;
		_mm_storel_epi64(&raw.packed, _mm_srli_epi16(_mm_add_epi16(_mm_mullo_epi16(vec_fg, vec_alpha), _mm_mullo_epi16(vec_bg, _mm_sub_epi16(vec_256, vec_alpha))), 8));
		return RendererApi::Color{
				static_cast<RendererApi::Color::component_compact_t>(raw.unpacked[0]),
				static_cast<RendererApi::Color::component_compact_t>(raw.unpacked[1]),
				static_cast<RendererApi::Color::component_compact_t>(raw.unpacked[2])
		};
#endif
#endif
	}
}